@@ -1,4 +1,4 @@
-;;;; org-test.el --- Tests for Org
+;;;; org-test.el --- Tests for Org  -*- lexical-binding: t; -*-
 
 ;; Copyright (c) 2010-2015 Sebastian Rose, Eric Schulte
 ;; Authors:
@@ -7,7 +7,7 @@
 ;;     David Maus, Brunswick, Germany, dmaus ictsoc de
 
 ;; Released under the GNU General Public License version 3
-;; see: http://www.gnu.org/licenses/gpl-3.0.html
+;; see: https://www.gnu.org/licenses/gpl-3.0.html
 
 ;; Definition of `special-mode' copied from Emacs23's simple.el to be
 ;; provide a testing environment for Emacs22.
@@ -31,6 +31,10 @@
 
 ;;;; Code:
 
+(require 'org)
+(require 'org-id)
+(require 'org-macs)
+
 ;;; Ob constants
 
 (defconst org-test-file-ob-anchor
@@ -75,7 +79,7 @@ org-test searches this directory up the directory tree.")
 (defconst org-test-dir
   (expand-file-name (file-name-directory (or load-file-name buffer-file-name))))
 
-(defconst org-base-dir
+(defconst org-test-base-dir
   (expand-file-name ".." org-test-dir))
 
 (defconst org-test-example-dir
@@ -87,17 +91,15 @@ org-test searches this directory up the directory tree.")
 (defconst org-test-no-heading-file
   (expand-file-name "no-heading.org" org-test-example-dir))
 
+(defconst org-test-attachments-file
+  (expand-file-name "attachments.org" org-test-example-dir))
+
 (defconst org-test-link-in-heading-file
   (expand-file-name "link-in-heading.org" org-test-dir))
 
-(defconst org-id-locations-file
-  (expand-file-name ".test-org-id-locations" org-test-dir))
-
 
 ;;; Functions for writing tests
-(put 'missing-test-dependency
-     'error-conditions
-     '(error missing-test-dependency))
+(define-error 'missing-test-dependency "org-test: Test dependency missing.")
 
 (defun org-test-for-executable (exe)
   "Throw an error if EXE is not available.
@@ -110,18 +112,18 @@ executable."
 	   exec-path :initial-value nil)
     (signal 'missing-test-dependency (list exe))))
 
-(defun org-test-buffer (&optional file)
+(defun org-test-buffer (&optional _file)
   "TODO:  Setup and return a buffer to work with.
 If file is non-nil insert its contents in there.")
 
-(defun org-test-compare-with-file (&optional file)
+(defun org-test-compare-with-file (&optional _file)
   "TODO:  Compare the contents of the test buffer with FILE.
 If file is not given, search for a file named after the test
 currently executed.")
 
 (defmacro org-test-at-id (id &rest body)
   "Run body after placing the point in the headline identified by ID."
-  (declare (indent 1))
+  (declare (indent 1) (debug t))
   `(let* ((id-location (org-id-find ,id))
 	  (id-file (car id-location))
 	  (visited-p (get-file-buffer id-file))
@@ -134,19 +136,19 @@ currently executed.")
 	     (condition-case nil
 		 (progn
 		   (org-show-subtree)
-		   (org-show-block-all))
+		   (org-show-all '(blocks)))
 	       (error nil))
 	     (save-restriction ,@body)))
        (unless (or visited-p (not to-be-removed))
 	 (kill-buffer to-be-removed)))))
-(def-edebug-spec org-test-at-id (form body))
 
 (defmacro org-test-in-example-file (file &rest body)
   "Execute body in the Org example file."
-  (declare (indent 1))
+  (declare (indent 1) (debug t))
   `(let* ((my-file (or ,file org-test-file))
 	  (visited-p (get-file-buffer my-file))
-	  to-be-removed)
+	  to-be-removed
+	  results)
      (save-window-excursion
        (save-match-data
 	 (find-file my-file)
@@ -158,31 +160,30 @@ currently executed.")
 	     (progn
 	       (outline-next-visible-heading 1)
 	       (org-show-subtree)
-	       (org-show-block-all))
+	       (org-show-all '(blocks)))
 	   (error nil))
-	 (save-restriction ,@body)))
+	 (setq results (save-restriction ,@body))))
      (unless visited-p
-       (kill-buffer to-be-removed))))
-(def-edebug-spec org-test-in-example-file (form body))
+       (kill-buffer to-be-removed))
+     results))
 
 (defmacro org-test-at-marker (file marker &rest body)
   "Run body after placing the point at MARKER in FILE.
 Note the uuidgen command-line command can be useful for
 generating unique markers for insertion as anchors into org
 files."
-  (declare (indent 2))
+  (declare (indent 2) (debug t))
   `(org-test-in-example-file ,file
      (goto-char (point-min))
      (re-search-forward (regexp-quote ,marker))
      ,@body))
-(def-edebug-spec org-test-at-marker (form form body))
 
 (defmacro org-test-with-temp-text (text &rest body)
   "Run body in a temporary buffer with Org mode as the active
 mode holding TEXT.  If the string \"<point>\" appears in TEXT
 then remove it and place the point there before running BODY,
 otherwise place the point at the beginning of the inserted text."
-  (declare (indent 1))
+  (declare (indent 1) (debug t))
   `(let ((inside-text (if (stringp ,text) ,text (eval ,text)))
 	 (org-mode-hook nil))
      (with-temp-buffer
@@ -194,33 +195,43 @@ otherwise place the point at the beginning of the inserted text."
 	       (goto-char (1+ (match-beginning 0))))
 	   (insert inside-text)
 	   (goto-char (point-min))))
+       (font-lock-ensure (point-min) (point-max))
        ,@body)))
-(def-edebug-spec org-test-with-temp-text (form body))
 
 (defmacro org-test-with-temp-text-in-file (text &rest body)
-  "Run body in a temporary file buffer with Org mode as the active mode."
-  (declare (indent 1))
-  (let ((results (cl-gensym)))
-    `(let ((file (make-temp-file "org-test"))
-	   (kill-buffer-query-functions nil)
-	   (inside-text (if (stringp ,text) ,text (eval ,text)))
-	   ,results)
-       (with-temp-file file (insert inside-text))
-       (find-file file)
-       (org-mode)
-       (setq ,results (progn ,@body))
-       (save-buffer) (kill-buffer (current-buffer))
-       (delete-file file)
-       ,results)))
-(def-edebug-spec org-test-with-temp-text-in-file (form body))
+  "Run body in a temporary file buffer with Org mode as the active mode.
+If the string \"<point>\" appears in TEXT then remove it and
+place the point there before running BODY, otherwise place the
+point at the beginning of the buffer."
+  (declare (indent 1) (debug t))
+  `(let ((file (make-temp-file "org-test"))
+	 (inside-text (if (stringp ,text) ,text (eval ,text)))
+	 buffer)
+     (with-temp-file file (insert inside-text))
+     (unwind-protect
+	 (progn
+	   ;; FIXME: For the rare cases where we do need to mess with windows,
+           ;; we should let `body' take care of displaying this buffer!
+	   (setq buffer (find-file file))
+	   (when (re-search-forward "<point>" nil t)
+	     (replace-match ""))
+	   (org-mode)
+	   (progn ,@body))
+       (let ((kill-buffer-query-functions nil))
+	 (when buffer
+	   (set-buffer buffer)
+	   ;; Ignore changes, we're deleting the file in the next step
+	   ;; anyways.
+	   (set-buffer-modified-p nil)
+	   (kill-buffer))
+	 (delete-file file)))))
 
-(defun org-test-table-target-expect (target &optional expect laps
-&rest tblfm)
+(defun org-test-table-target-expect (target &optional expect laps &rest tblfm)
   "For all TBLFM: Apply the formula to TARGET, compare EXPECT with result.
 Either LAPS and TBLFM are nil and the table will only be aligned
 or LAPS is the count of recalculations that should be made on
 each TBLFM.  To save ERT run time keep LAPS as low as possible to
-get the table stable.  Anyhow, if LAPS is 'iterate then iterate,
+get the table stable.  Anyhow, if LAPS is `iterate' then iterate,
 but this will run one recalculation longer.  When EXPECT is nil
 it will be set to TARGET.
 
@@ -268,18 +279,57 @@ setting `pp-escape-newlines' to nil manually."
 	;; on multiple lines in the ERT results buffer.
 	(setq pp-escape-newlines back)))))
 
+(defun org-test-with-tramp-remote-dir--worker (body)
+  "Worker for `org-test-with-tramp-remote-dir'."
+  (let ((env-def (getenv "REMOTE_TEMPORARY_FILE_DIRECTORY")))
+    (cond
+     (env-def (funcall body env-def))
+     ((eq system-type 'windows-nt) (funcall body null-device))
+     (t (require 'tramp)
+        (defvar tramp-methods)
+        (defvar tramp-default-host-alist)
+        (let ((tramp-methods
+               (cons '("mock"
+                       (tramp-login-program        "sh")
+                       (tramp-login-args           (("-i")))
+                       (tramp-remote-shell         "/bin/sh")
+                       (tramp-remote-shell-args    ("-c"))
+                       (tramp-connection-timeout   10))
+                     tramp-methods))
+              (tramp-default-host-alist
+               `(("\\`mock\\'" nil ,(system-name)))))
+          (funcall body (format "/mock::%s" temporary-file-directory)))))))
+
+(defmacro org-test-with-tramp-remote-dir (dir &rest body)
+  "Bind the symbol DIR to a remote directory and execute BODY.
+Return the value of the last form in BODY.  The directory DIR
+will be something like \"/mock::/tmp/\", which allows to test
+Tramp related features.  We mostly follow
+`tramp-test-temporary-file-directory' from GNU Emacs tests."
+  (declare (debug (sexp body)) (indent 2))
+  `(org-test-with-tramp-remote-dir--worker (lambda (,dir) ,@body)))
+
+
 
 ;;; Navigation Functions
-(when (featurep 'jump)
+
+(defmacro org--compile-when (test &rest body)
+  (declare (debug t) (indent 1))
+  (let ((exp `(progn ,@body)))
+    (if (eval test t)
+        exp
+      `(when ,test (eval ',exp t)))))
+
+(org--compile-when (featurep 'jump)
   (defjump org-test-jump
     (("lisp/\\1.el" . "testing/lisp/test-\\1.el")
      ("lisp/\\1.el" . "testing/lisp/\\1.el/test.*.el")
      ("testing/lisp/test-\\1.el" . "lisp/\\1.el")
      ("testing/lisp/\\1.el" . "lisp/\\1.el/test.*.el"))
-    (concat org-base-dir "/")
+    (concat org-test-base-dir "/")
     "Jump between Org files and their tests."
     (lambda (path)
-      (let* ((full-path (expand-file-name path org-base-dir))
+      (let* ((full-path (expand-file-name path org-test-base-dir))
 	     (file-name (file-name-nondirectory path))
 	     (name (file-name-sans-extension file-name)))
 	(find-file full-path)
@@ -289,7 +339,7 @@ setting `pp-escape-newlines' to nil manually."
 	 " " user-full-name "\n"
 	 ";; Authors: " user-full-name "\n\n"
 	 ";; Released under the GNU General Public License version 3\n"
-	 ";; see: http://www.gnu.org/licenses/gpl-3.0.html\n\n"
+	 ";; see: https://www.gnu.org/licenses/gpl-3.0.html\n\n"
 	 ";;;; Comments:\n\n"
 	 ";; Template test file for Org tests\n\n"
 	 "\n"
@@ -307,10 +357,11 @@ setting `pp-escape-newlines' to nil manually."
 	 "  (should-not nil)\n"
 	 "  (should-error (error \"errr...\")))\n\n\n"
 	 "(provide '" name ")\n\n"
-	 ";;; " file-name " ends here\n") full-path))
+	 ";;; " file-name " ends here\n")
+	full-path))
     (lambda () ((lambda (res) (if (listp res) (car res) res)) (which-function)))))
 
-(define-key emacs-lisp-mode-map "\M-\C-j" 'org-test-jump)
+(define-key emacs-lisp-mode-map "\M-\C-j" #'org-test-jump)
 
 
 ;;; Miscellaneous helper functions
@@ -336,28 +387,34 @@ setting `pp-escape-newlines' to nil manually."
 (defun org-test-load ()
   "Load up the Org test suite."
   (interactive)
-  (cl-flet ((rld (base)
-	      ;; Recursively load all files, if files throw errors
-	      ;; then silently ignore the error and continue to the
-	      ;; next file.  This allows files to error out if
-	      ;; required executables aren't available.
-	      (mapc
-	       (lambda (path)
-		 (if (file-directory-p path)
-		     (rld path)
-		   (condition-case err
-		       (when (string-match "^[A-Za-z].*\\.el$"
-					   (file-name-nondirectory path))
-			 (load-file path))
-		     (missing-test-dependency
-		      (let ((name (intern
-				   (concat "org-missing-dependency/"
-					   (file-name-nondirectory
-					    (file-name-sans-extension path))))))
-			(eval `(ert-deftest ,name ()
-				 :expected-result :failed (should nil))))))))
-	       (directory-files base 'full
-				"^\\([^.]\\|\\.\\([^.]\\|\\..\\)\\).*\\.el$"))))
+  (cl-flet
+      ((rld (base)
+	 ;; Recursively load all files, if files throw errors
+	 ;; then silently ignore the error and continue to the
+	 ;; next file.  This allows files to error out if
+	 ;; required executables aren't available.
+	 (mapc
+	  (lambda (path)
+	    (if (file-directory-p path)
+		(rld path)
+	      (condition-case nil
+		  (when (string-match "\\`[A-Za-z].*\\.el\\'"
+				      (file-name-nondirectory path))
+                    (let ((feature-name
+                           (intern
+                            (file-name-base
+                             (file-name-nondirectory path)))))
+		      (require feature-name path)))
+		(missing-test-dependency
+		 (let ((name (intern
+			      (concat "org-missing-dependency/"
+				      (file-name-nondirectory
+				       (file-name-sans-extension path))))))
+		   (eval `(ert-deftest ,name ()
+                            (skip-unless nil) ;; Make it prominent.
+			    :expected-result :failed (should nil))))))))
+	  (directory-files base 'full
+			   "\\`\\([^.]\\|\\.\\([^.]\\|\\..\\)\\).*\\.el\\'"))))
     (rld (expand-file-name "lisp" org-test-dir))))
 
 (defun org-test-current-defun ()
@@ -389,6 +446,8 @@ setting `pp-escape-newlines' to nil manually."
       (when (buffer-live-p b) (kill-buffer b)))))
 
 (defun org-test-update-id-locations ()
+  (setq org-id-locations-file
+        (expand-file-name ".test-org-id-locations" org-test-dir))
   (org-id-update-id-locations
    (directory-files
     org-test-example-dir 'full
@@ -401,7 +460,9 @@ Load all test files first."
   (let ((org-id-track-globally t)
 	(org-test-selector
 	 (if org-test-selector org-test-selector "\\(org\\|ob\\)"))
-	org-confirm-babel-evaluate org-startup-folded vc-handled-backends)
+	org-confirm-babel-evaluate org-startup-folded vc-handled-backends
+        ;; Catch errors in diary sexps better.
+        (calendar-debug-sexp t))
     (org-test-touch-all-examples)
     (org-test-update-id-locations)
     (org-test-load)
@@ -415,9 +476,132 @@ Load all test files first."
   (org-test-touch-all-examples)
   (org-test-update-id-locations)
   (org-test-load)
-  (ert "\\(org\\|ob\\)")
+  (let (;; Catch errors in diary sexps better.
+        (calendar-debug-sexp t))
+    (ert "\\(org\\|ob\\)"))
   (org-test-kill-all-examples))
 
+(defmacro org-test-at-time (time &rest body)
+  "Run body while pretending that the current time is TIME.
+TIME can be a non-nil Lisp time value, or a string specifying a date and time."
+  (declare (indent 1))
+  (let ((tm (cl-gensym))
+	(at (cl-gensym)))
+    `(let* ((,tm ,time)
+	    (,at (if (stringp ,tm)
+		     (org-time-string-to-time ,tm)
+		   ,tm)))
+       (cl-letf
+	   ;; Wrap builtins whose behavior can depend on the current time.
+	   (((symbol-function 'current-time)
+	     (lambda () ,at))
+	    ((symbol-function 'current-time-string)
+	     (lambda (&optional time &rest args)
+	       (apply ,(symbol-function 'current-time-string)
+		      (or time ,at) args)))
+	    ((symbol-function 'current-time-zone)
+	     (lambda (&optional time &rest args)
+	       (apply ,(symbol-function 'current-time-zone)
+		      (or time ,at) args)))
+	    ((symbol-function 'decode-time)
+	     (lambda (&optional time zone form)
+               (condition-case nil
+                   (funcall ,(symbol-function 'decode-time)
+			    (or time ,at) zone form)
+                 (wrong-number-of-arguments
+                  (funcall ,(symbol-function 'decode-time)
+			   (or time ,at))))))
+	    ((symbol-function 'encode-time)
+	     (lambda (time &rest args)
+	       (apply ,(symbol-function 'encode-time) (or time ,at) args)))
+	    ((symbol-function 'float-time)
+	     (lambda (&optional time)
+	       (funcall ,(symbol-function 'float-time) (or time ,at))))
+	    ((symbol-function 'format-time-string)
+	     (lambda (format &optional time &rest args)
+	       (apply ,(symbol-function 'format-time-string)
+		      format (or time ,at) args)))
+	    ((symbol-function 'set-file-times)
+	     (lambda (file &optional time)
+	       (funcall ,(symbol-function 'set-file-times) file (or time ,at))))
+	    ((symbol-function 'time-add)
+	     (lambda (a b) (funcall ,(symbol-function 'time-add)
+			       (or a ,at) (or b ,at))))
+	    ((symbol-function 'time-equal-p)
+	     (lambda (a b) (funcall ,(symbol-function 'time-equal-p)
+				    (or a ,at) (or b ,at))))
+	    ((symbol-function 'time-less-p)
+	     (lambda (a b) (funcall ,(symbol-function 'time-less-p)
+				    (or a ,at) (or b ,at))))
+	    ((symbol-function 'time-subtract)
+	     (lambda (a b) (funcall ,(symbol-function 'time-subtract)
+				    (or a ,at) (or b ,at)))))
+	 ,@body))))
+
+(defmacro org-test-capture-warnings (&rest body)
+  "Capture all warnings passed to `org-display-warning' within BODY."
+  (declare (indent 0) (debug t))
+  `(let ((messages (list)))
+     (cl-letf (((symbol-function 'org-display-warning)
+                (lambda (message)
+                  (setq messages (cons message messages)))))
+       ,@body)
+     (nreverse messages)))
+
+(defconst org-test-day-of-weeks-seconds
+  [302400                               ; Sun
+   388800                               ; Mon
+   475200                               ; Tue
+   561600                               ; Wed
+   648000                               ; Thu
+   734400                               ; Fri
+   820800]                              ; Sat
+  "Epoch seconds for generating days of week strings.
+Starts at Sunday, ends at Saturday.")
+
+(defconst org-test-day-of-weeks-abbrev
+  (apply #'vector
+         (seq-map (lambda (s) (format-time-string "%a" s t))
+                  org-test-day-of-weeks-seconds))
+  "Vector of abbreviated names of days of week.
+See `org-test-day-of-weeks-seconds'.")
+
+(defconst org-test-day-of-weeks-full
+  (apply #'vector
+         (seq-map (lambda (s) (format-time-string "%A" s t))
+                  org-test-day-of-weeks-seconds))
+  "Vector of full names for days of week.
+See `org-test-day-of-weeks-seconds'.")
+
+(defun org-test-get-day-name (day &optional full)
+  "Return string containing locale-specific DAY abbrev.
+DAY Should be Mon, Tue, ...
+When FULL is non-nil, return the full name like Monday, Tuesday, ..."
+  (aref
+   (if full org-test-day-of-weeks-full
+     org-test-day-of-weeks-abbrev)
+   (pcase day
+     ((or "Sun" "Sunday") 0)
+     ((or "Mon" "Monday") 1)
+     ((or "Tue" "Tuesday") 2)
+     ((or "Wed" "Wednesday") 3)
+     ((or "Thu" "Thursday") 4)
+     ((or "Fri" "Friday") 5)
+     ((or "Sat" "Saturday") 6)
+     (_ (error "Unknown day of week: %s" day)))))
+
+(defmacro org-test-with-exported-text (backend source &rest body)
+  "Run BODY in export buffer for SOURCE string via BACKEND."
+  (declare (indent 2))
+  `(org-test-with-temp-text ,source
+     (let ((export-buffer (generate-new-buffer "Org temporary export")))
+       (unwind-protect
+           (progn
+             (org-export-to-buffer ,backend export-buffer)
+             (with-current-buffer export-buffer
+               ,@body))
+         (kill-buffer export-buffer)))))
+
 (provide 'org-test)
 
 ;;; org-test.el ends here
